/*
**             Copyright 2023 by Kvaser AB, Molndal, Sweden
**                         http://www.kvaser.com
**
** This software is dual licensed under the following two licenses:
** BSD-new and GPLv2. You may use either one. See the included
** COPYING file for details.
**
** License: BSD-new
** ==============================================================================
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the <organization> nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
** AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
** ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
** LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
** CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
** SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
** BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
** IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
** ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
** POSSIBILITY OF SUCH DAMAGE.
**
**
** License: GPLv2
** ==============================================================================
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
**
**
** IMPORTANT NOTICE:
** ==============================================================================
** This source code is made available for free, as an open license, by Kvaser AB,
** for use with its applications. Kvaser AB does not accept any liability
** whatsoever for any third party patent or other immaterial property rights
** violations that may result from any usage of this source code, regardless of
** the combination of source code and various applications that it can be used
** in, or with.
**
** -----------------------------------------------------------------------------
*/

/**
 * Kvaser Smartfusion2 SPI flash implementation
 * Compatible with IS25LP016 SPI flash
 *
 * @note: This module is NOT thread safe, only perform
 *        one ongoing operation at any time!
 *
 */
#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 11, 0)
#include <linux/kernel.h> /* DIV_ROUND_CLOSEST() */
#else
#include <linux/math.h> /* DIV_ROUND_CLOSEST() */
#endif /* LINUX_VERSION_CODE < 5.11 */

#include <linux/slab.h>
#include "debugprint.h"
#include "k_assert.h"
#include "util.h"
#include "spi_flash.h"

#define SIMULATE_WRITE 0

#define JEDEC_IS25LP016 0x009D6015

#define READ_ARRAY_OPCODE 0x03
#define DEVICE_ID_READ    0x90

#define WRITE_ENABLE_CMD       0x06
#define WRITE_DISABLE_CMD      0x04
#define PROGRAM_PAGE_CMD       0x02
#define WRITE_STATUS1_OPCODE   0x01
#define CHIP_ERASE_OPCODE      0x60
#define ERASE_4K_BLOCK_OPCODE  0x20
#define ERASE_32K_BLOCK_OPCODE 0x52
#define ERASE_64K_BLOCK_OPCODE 0xD8
#define READ_STATUS            0x05
#define CMD_READ_JEDEC_ID_SPI  0x9F /* Read JEDEC id in SPI mode*/

#define BUSY_BIT_MASK 0x01

#define UNPROTECT_SECTOR_OPCODE 0x39

#define CLK_DIV 8U

/*
 * Number of bytes per page in the flash device.
 */
#define FLASH_PAGE_SIZE 256U

/**
 * @brief Allocate a SPI flash handle
 */
int SPI_FLASH_sf2_alloc_handle(struct spi_flash *spif)
{
    K_ASSERT(spif != NULL);
    K_ASSERT(spif->hnd == NULL);

    spif->hnd = kzalloc(sizeof(u8), GFP_KERNEL);

    if (spif->hnd == NULL)
        return SPI_FLASH_STATUS_MALLOC;
    else
        return SPI_FLASH_STATUS_SUCCESS;
}

/**
 * @brief Free a SPI flash handle
 */
void SPI_FLASH_sf2_free_handle(struct spi_flash *spif)
{
    K_ASSERT(spif != NULL);

    kfree(spif->hnd);
}

/**
 * @brief Initialize and configure SPI comm.
 */
int SPI_FLASH_sf2_init(struct spi_flash *spif, void __iomem *base_addr)
{
    K_ASSERT(spif != NULL);
    K_ASSERT(spif->hnd != NULL);
    K_ASSERT(base_addr != NULL);

    return SPI_FLASH_STATUS_SUCCESS;
}

/**
 * @brief Dummy function for API compatibility
 */
int SPI_FLASH_sf2_start(struct spi_flash *spif)
{
    NOT_USED(spif);
    return SPI_FLASH_STATUS_SUCCESS;
}

/**
 * @brief Dummy function for API compatibility
 */
int SPI_FLASH_sf2_stop(struct spi_flash *spif)
{
    NOT_USED(spif);
    return SPI_FLASH_STATUS_SUCCESS;
}

/**
 * @brief Read the status register of the Flash.
 */
int SPI_FLASH_sf2_get_status(struct spi_flash *spif, u8 *result)
{
    return SPI_FLASH_STATUS_SUCCESS;
}

/**
 * @brief Wait until the serial flash is ready to accept a command
 */
int SPI_FLASH_sf2_wait_ready(struct spi_flash *spif, u32 timeout_ms)
{
    return SPI_FLASH_STATUS_SUCCESS;
}

/**
 * @brief Enable writes to the serial flash memory.
 */
int SPI_FLASH_sf2_write_enable(struct spi_flash *spif)
{
    return SPI_FLASH_STATUS_SUCCESS;
}

/**
 * @brief Disable writes to the serial flash memory.
 */
int SPI_FLASH_sf2_write_disable(struct spi_flash *spif)
{
    return SPI_FLASH_STATUS_SUCCESS;
}

/**
 * @brief Get JEDEC id
 */
int SPI_FLASH_sf2_get_jedec(struct spi_flash *spif, u32 *jedec)
{
    *jedec = JEDEC_IS25LP016;
    return SPI_FLASH_STATUS_SUCCESS;
}

/**
 * @brief Verify flash JEDEC id
 */
bool SPI_FLASH_sf2_verify_jedec(struct spi_flash *spif)
{
    int status;
    u32 jedec = 0;

    status = SPI_FLASH_sf2_get_jedec(spif, &jedec);
    if (status != SPI_FLASH_STATUS_SUCCESS) {
        DEBUGPRINT(1, "Reading JEDEC id failed with status %i", status);
        return false;
    }

    DEBUGPRINT(3, "JEDEC id = 0x%08X", jedec);
    return (jedec == JEDEC_IS25LP016);
}

/**
 * @brief Erase a 64 kiB block
 */
int SPI_FLASH_sf2_erase_64K(struct spi_flash *spif, u32 addr, u32 timeout_ms)
{
    return SPI_FLASH_STATUS_SUCCESS;
}

/**
 * @brief Erase one or more 64 kiB blocks
 */
int SPI_FLASH_sf2_erase_multi_64K(struct spi_flash *spif, u32 addr, u32 num_bytes, u32 timeout_ms)
{
    return SPI_FLASH_STATUS_SUCCESS;
}

/**
 * @brief Write up to 256 bytes of data to a specified location in the serial flash memory.
 */
int SPI_FLASH_sf2_write_page(struct spi_flash *spif, u32 addr, const u8 *buf, u32 num_bytes)
{
    return SPI_FLASH_STATUS_SUCCESS;
}

/**
 * @brief Write multiple pages of up to 256 bytes of data to a specified
 *        location in the serial flash memory.
 */
int SPI_FLASH_sf2_write_multi_page(struct spi_flash *spif, u32 addr, const u8 *buf, u32 num_bytes)
{
    return SPI_FLASH_STATUS_SUCCESS;
}

/**
* @brief Read multiple pages of up to 256 bytes of data from a
*        specified location in the serial flash memory.
*/
int SPI_FLASH_sf2_read(struct spi_flash *spif, u32 addr, u8 *buf, u32 num_bytes)
{
    return SPI_FLASH_STATUS_SUCCESS;
}

/**
* @brief Compare the contents of a buffer with the contents of a
*        specified location in the serial flash memory.
*/
int SPI_FLASH_sf2_compare(struct spi_flash *spif, u32 addr, const u8 *buf, u32 num_bytes)
{
    return SPI_FLASH_STATUS_SUCCESS;
}

const struct SPI_FLASH_ops SPI_FLASH_sf2_ops = {
    .alloc_handle = SPI_FLASH_sf2_alloc_handle,
    .free_handle = SPI_FLASH_sf2_free_handle,
    .init = SPI_FLASH_sf2_init,
    .start = SPI_FLASH_sf2_start,
    .stop = SPI_FLASH_sf2_stop,
    .get_status = SPI_FLASH_sf2_get_status,
    .wait_ready = SPI_FLASH_sf2_wait_ready,
    .write_enable = SPI_FLASH_sf2_write_enable,
    .write_disable = SPI_FLASH_sf2_write_disable,
    .get_jedec = SPI_FLASH_sf2_get_jedec,
    .verify_jedec = SPI_FLASH_sf2_verify_jedec,
    .erase_64K = SPI_FLASH_sf2_erase_64K,
    .erase_multi_64K = SPI_FLASH_sf2_erase_multi_64K,
    .write_page = SPI_FLASH_sf2_write_page,
    .write_multi_page = SPI_FLASH_sf2_write_multi_page,
    .read = SPI_FLASH_sf2_read,
    .compare = SPI_FLASH_sf2_compare,
};
